﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Web.Http;
using System.Web.Http.ValueProviders;
using System.Text;
using BookStore.Services.Models;
using BookStore.DataLayer;
using BookStore.Models;
using BlogSystem.Services.Attributes;

namespace BookStore.Services.Controllers
{
    public class UsersController : BaseApiController
    {
        private const int MinUsernameLength = 6;
        private const int MaxUsernameLength = 30;
        private const string ValidUsernameCharacters =
            "qwertyuioplkjhgfdsazxcvbnmQWERTYUIOPLKJHGFDSAZXCVBNM1234567890_.";

        private const string SessionKeyChars =
            "qwertyuioplkjhgfdsazxcvbnmQWERTYUIOPLKJHGFDSAZXCVBNM";
        private static readonly Random rand = new Random();

        private const int SessionKeyLength = 50;

        private const int Sha1Length = 40;

        /*
        {  "username": "DonchoMinkov",
           "nickname": "Doncho Minkov",
           "authCode":   "bfff2dd4f1b310eb0dbf593bd83f94dd8d34077e" }

{
"username": "ssssss",
"displayName": "SSSSSS",
"authCode": "bfff2dd4f1d310eb0dbf593bd83f94dd8d34077s"
}
        */

        [ActionName("all")]
        public IQueryable<UserModelFull> GetAll(
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(() =>
            {
                var context = new BookstoreContext();

                var user = context.Users.FirstOrDefault(
                    usr => usr.SessionKey == sessionKey);

                if (user == null)
                {
                    throw new InvalidOperationException("Invalid sessionKey.");
                }

                if (!user.IsAdmin)
                {
                    throw new InvalidOperationException("Access Denied. You do not have permissions.");
                }

                var userEntities = context.Users;
                var models =
                    (from userEntity in userEntities
                     select new UserModelFull()
                     {
                         Id = userEntity.Id,
                         Username = userEntity.Username,
                         AuthCode = userEntity.AuthCode,
                         Balance = userEntity.Balance,
                         IsAdmin = userEntity.IsAdmin
                     });

                return models;
            });

            return responseMsg;
        }

        [ActionName("details")]
        public UserModelFull GetDetails(int userId,
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(() =>
            {
                var context = new BookstoreContext();

                var user = context.Users.FirstOrDefault(
                    usr => usr.SessionKey == sessionKey);
                if (user == null || !user.IsAdmin)
                {
                    throw new InvalidOperationException("Invalid sessionKey");
                }

                var userEntities = context.Users;
                var model = userEntities.FirstOrDefault(x=> x.Id == userId);

                return new UserModelFull()
                     {
                         Id = model.Id,
                         Username = model.Username,
                         AuthCode = model.AuthCode,
                         Balance = model.Balance,
                         IsAdmin = model.IsAdmin
                     };
            });

            return responseMsg;
        }

        [ActionName("access")]
        public HttpResponseMessage PutDetails([FromBody] UserModelFull userDetails,
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(() =>
            {
                var context = new BookstoreContext();

                var user = context.Users.FirstOrDefault(
                    usr => usr.SessionKey == sessionKey);

                if (user == null || !user.IsAdmin)
                {
                    throw new InvalidOperationException("Invalid sessionKey");
                }

                var userEntities = context.Users;

                if (userDetails == null || userDetails.IsAdmin == null || userDetails.Id == null)
                {
                    throw new InvalidOperationException("Invalid data sent.");
                }

                var model = userEntities.FirstOrDefault(x => x.Id == userDetails.Id);

                if (model == null)
                {
                    throw new InvalidOperationException("No such user.");
                }

                model.IsAdmin = userDetails.IsAdmin;
                context.SaveChanges();

                var response = this.Request.CreateResponse(HttpStatusCode.OK);
                
                return response;
            });

            return responseMsg;
        }

        [ActionName("profile")]
        public UserModelProfile GetProfile(
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(() =>
            {
                var context = new BookstoreContext();

                var user = context.Users.FirstOrDefault(
                    usr => usr.SessionKey == sessionKey);
                if (user == null)
                {
                    throw new InvalidOperationException("Invalid sessionKey");
                }

                var model = new UserModelProfile()
                {
                    Username = user.Username,
                    AuthCode = user.AuthCode,
                    Balance = user.Balance,
                    IsAdmin = user.IsAdmin,
                    Records = (from records in user.BookRecords
                                     select new BookRecordModel
                                     {
                                         Username = user.Username,
                                         RentDate = records.RentDate,
                                         ReturnedDate = records.ReturnedDate,
                                         Book = (from bookEntity in context.Books
                                                where bookEntity.Id == records.BookId
                                                 select new BookModel()
                                                 {
                                                     Id = bookEntity.Id,
                                                     Title = bookEntity.Title,
                                                     Author = bookEntity.Author,
                                                     Store = bookEntity.Store.StoreName,
                                                     Genre = bookEntity.Genre,
                                                     RentCost = bookEntity.RentCost,
                                                     Pages = bookEntity.Pages,
                                                     IsAvailable = bookEntity.IsAvailable ? "True" : "False"
                                                 }).FirstOrDefault()
                                    })
                };

                return model;
            });

            return responseMsg;
        }

        [ActionName("deposit")]
        public HttpResponseMessage PutDeposit(decimal amount,
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(
              () =>
              {
                  var context = new BookstoreContext();
                  using (context)
                  {
                      var user = context.Users.FirstOrDefault(
                          usr => usr.SessionKey == sessionKey);

                      if (user == null)
                      {
                          throw new InvalidOperationException("Invalid sessionKey");
                      }

                      if (amount < 0)
                      {
                          throw new InvalidOperationException("Invalid amount");
                      }

                      user.Balance += amount;
                      context.SaveChanges();

                      var response =
                          this.Request.CreateResponse(HttpStatusCode.OK);
                      return response;
                  }
              });

            return responseMsg;
        }

        [ActionName("register")]
        public HttpResponseMessage PostRegisterUser(UserModel model)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(
                () =>
                {
                    var context = new BookstoreContext();
                    using (context)
                    {
                        this.ValidateUsername(model.Username);
                        this.ValidateAuthCode(model.AuthCode);
                        var usernameToLower = model.Username.ToLower();
                        var user = context.Users.FirstOrDefault(
                            usr => usr.Username == usernameToLower);

                        if (user != null)
                        {
                            throw new InvalidOperationException("Users already exists");
                        }

                        user = new User()
                        {
                            Username = usernameToLower,
                            AuthCode = model.AuthCode
                        };

                        context.Users.Add(user);
                        context.SaveChanges();

                        user.SessionKey = user.Id.ToString();//this.GenerateSessionKey(user.Id);
                        context.SaveChanges();

                        var loggedModel = new LoggedUserModel()
                        {
                            Username = user.Username,
                            SessionKey = user.SessionKey
                        };

                        var response =
                            this.Request.CreateResponse(HttpStatusCode.Created,
                                            loggedModel);
                        return response;
                    }
                });

            return responseMsg;
        }

        [ActionName("login")]
        public HttpResponseMessage PostLoginUser(UserModel model)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(
              () =>
              {
                  var context = new BookstoreContext();
                  using (context)
                  {
                      this.ValidateUsername(model.Username);
                      this.ValidateAuthCode(model.AuthCode);
                      var usernameToLower = model.Username.ToLower();
                      var user = context.Users.FirstOrDefault(
                          usr => usr.Username == usernameToLower
                          && usr.AuthCode == model.AuthCode);

                      if (user == null)
                      {
                          throw new InvalidOperationException("Invalid username or password");
                      }

                      if (user.SessionKey == null)
                      {
                          user.SessionKey = user.Id.ToString();//this.GenerateSessionKey(user.Id);
                          context.SaveChanges();
                      }

                      var loggedModel = new LoggedUserModel()
                      {
                          Username = user.Username,
                          SessionKey = user.SessionKey
                      };

                      var response =
                          this.Request.CreateResponse(HttpStatusCode.Created,
                                          loggedModel);
                      return response;
                  }
              });

            return responseMsg;
        }

        [ActionName("logout")]
        public HttpResponseMessage PutLogoutUser(
            [ValueProvider(typeof(HeaderValueProviderFactory<string>))] string sessionKey)
        {
            var responseMsg = this.PerformOperationAndHandleExceptions(
              () =>
              {
                  var context = new BookstoreContext();
                  using (context)
                  {
                      var user = context.Users.FirstOrDefault(
                          usr => usr.SessionKey == sessionKey);

                      if (user == null)
                      {
                          throw new InvalidOperationException("Invalid sessionKey");
                      }

                      user.SessionKey = null;
                      context.SaveChanges();

                      var response =
                          this.Request.CreateResponse(HttpStatusCode.OK);
                      return response;
                  }
              });

            return responseMsg;
        }

        private string GenerateSessionKey(int userId)
        {
            StringBuilder skeyBuilder = new StringBuilder(SessionKeyLength);
            skeyBuilder.Append(userId);
            while (skeyBuilder.Length < SessionKeyLength)
            {
                var index = rand.Next(SessionKeyChars.Length);
                skeyBuilder.Append(SessionKeyChars[index]);
            }
            return skeyBuilder.ToString();
        }

        private void ValidateAuthCode(string authCode)
        {
            if (authCode == null || authCode.Length != Sha1Length)
            {
                throw new ArgumentOutOfRangeException("Password should be encrypted");
            }
        }

        private void ValidateUsername(string username)
        {
            if (username == null)
            {
                throw new ArgumentNullException("Username cannot be null");
            }
            else if (username.Length < MinUsernameLength)
            {
                throw new ArgumentOutOfRangeException(
                    string.Format("Username must be at least {0} characters long",
                    MinUsernameLength));
            }
            else if (username.Length > MaxUsernameLength)
            {
                throw new ArgumentOutOfRangeException(
                    string.Format("Username must be less than {0} characters long",
                    MaxUsernameLength));
            }
            else if (username.Any(ch => !ValidUsernameCharacters.Contains(ch)))
            {
                throw new ArgumentOutOfRangeException(
                    "Username must contain only Latin letters, digits .,_");
            }
        }
    }
}
